Index: ip_id.c
===================================================================
--- ip_id.c	(revision 280788)
+++ ip_id.c	(working copy)
@@ -74,9 +74,10 @@
  * enabled.
  */
 
-#include <sys/types.h>
+#include <sys/param.h>
+#include <sys/systm.h>
+#include <sys/counter.h>
 #include <sys/malloc.h>
-#include <sys/param.h>
 #include <sys/time.h>
 #include <sys/kernel.h>
 #include <sys/libkern.h>
@@ -83,7 +84,7 @@
 #include <sys/lock.h>
 #include <sys/mutex.h>
 #include <sys/random.h>
-#include <sys/systm.h>
+#include <sys/smp.h>
 #include <sys/sysctl.h>
 #include <sys/bitstring.h>
 
@@ -92,8 +93,10 @@
 #include <netinet/in.h>
 #include <netinet/ip_var.h>
 
+/*
+ * Random ID state engine.
+ */
 static MALLOC_DEFINE(M_IPID, "ipid", "randomized ip id state");
-
 static VNET_DEFINE(uint16_t *, id_array);
 static VNET_DEFINE(bitstr_t *, id_bits);
 static VNET_DEFINE(int, array_ptr);
@@ -109,6 +112,12 @@
 #define	V_random_id_total	VNET(random_id_total)
 #define	V_ip_id_mtx	VNET(ip_id_mtx)
 
+/*
+ * Non-random ID state engine is simply a per-cpu counter.
+ */
+static VNET_DEFINE(counter_u64_t, ip_id);
+#define	V_ip_id		VNET(ip_id)
+
 static void	ip_initid(int);
 static int	sysctl_ip_id_change(SYSCTL_HANDLER_ARGS);
 static void	ipid_sysinit(void);
@@ -191,6 +200,14 @@
 	return (new_id);
 }
 
+uint16_t
+ip_newid(void)
+{
+
+	counter_u64_add(V_ip_id, 1);
+	return (htons((*(uint64_t *)zpcpu_get(V_ip_id)) & 0xffff));
+}
+
 static void
 ipid_sysinit(void)
 {
@@ -197,6 +214,9 @@
 
 	mtx_init(&V_ip_id_mtx, "ip_id_mtx", NULL, MTX_DEF);
 	ip_initid(8192);
+	V_ip_id = counter_u64_alloc(M_WAITOK);
+	for (int i = 0; i < mp_ncpus; i++)
+		arc4rand(zpcpu_get_cpu(V_ip_id, i), sizeof(uint64_t), 0);
 }
 VNET_SYSINIT(ip_id, SI_SUB_PROTO_DOMAIN, SI_ORDER_ANY, ipid_sysinit, NULL);
 
@@ -207,5 +227,6 @@
 	mtx_destroy(&V_ip_id_mtx);
 	free(V_id_array, M_IPID);
 	free(V_id_bits, M_IPID);
+	counter_u64_free(V_ip_id);
 }
 VNET_SYSUNINIT(ip_id, SI_SUB_PROTO_DOMAIN, SI_ORDER_ANY, ipid_sysuninit, NULL);
Index: ip_input.c
===================================================================
--- ip_input.c	(revision 280779)
+++ ip_input.c	(working copy)
@@ -331,8 +331,6 @@
 	struct protosw *pr;
 	int i;
 
-	V_ip_id = time_second & 0xffff;
-
 	TAILQ_INIT(&V_in_ifaddrhead);
 	V_in_ifaddrhashtbl = hashinit(INADDR_NHASH, M_IFADDR, &V_in_ifaddrhmask);
 
Index: ip_output.c
===================================================================
--- ip_output.c	(revision 280779)
+++ ip_output.c	(working copy)
@@ -91,8 +91,6 @@
 
 #include <security/mac/mac_framework.h>
 
-VNET_DEFINE(uint32_t, ip_id);
-
 #ifdef MBUF_STRESS_TEST
 static int mbuf_frag_size = 0;
 SYSCTL_INT(_net_inet_ip, OID_AUTO, mbuf_frag_size, CTLFLAG_RW,
Index: ip_var.h
===================================================================
--- ip_var.h	(revision 280779)
+++ ip_var.h	(working copy)
@@ -174,7 +174,6 @@
 struct route;
 struct sockopt;
 
-VNET_DECLARE(uint32_t, ip_id);			/* ip packet ctr, for ids */
 VNET_DECLARE(int, ip_defttl);			/* default IP ttl */
 VNET_DECLARE(int, ipforwarding);		/* ip forwarding */
 #ifdef IPSTEALTH
@@ -229,6 +228,7 @@
 	    struct mbuf *);
 void	ip_slowtimo(void);
 uint16_t	ip_randomid(void);
+uint16_t	ip_newid(void);
 int	rip_ctloutput(struct socket *, struct sockopt *);
 void	rip_ctlinput(int, struct sockaddr *, void *);
 void	rip_init(void);
@@ -305,19 +305,7 @@
 
 VNET_DECLARE(int, ip_do_randomid);
 #define	V_ip_do_randomid	VNET(ip_do_randomid)
-static __inline uint16_t
-ip_newid(void)
-{
-	uint16_t res;
 
-	if (V_ip_do_randomid != 0)
-		return (ip_randomid());
-	else {
-		res = atomic_fetchadd_32(&V_ip_id, 1) & 0xFFFF;
-		return (htons(res));
-	}
-}
-
 #endif /* _KERNEL */
 
 #endif /* !_NETINET_IP_VAR_H_ */
